/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// gui_cursor.c
//

#include "gui_local.h"

/*
=============================================================================

	BOUNDS CALCULATION

=============================================================================
*/

/*
==================
GUI_GenerateBounds
==================
*/
static void GUI_GenerateBounds_r (guiWindow_t *window)
{
	guiWindow_t	*child;
	uint32		i;

	// Generate the window bounds
	window->mins[0] = window->xPos;
	window->mins[1] = window->yPos;
	window->maxs[0] = window->mins[0] + window->width;
	window->maxs[1] = window->mins[1] + window->height;

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++) {
		if (!child->visible || child->flags & WNFL_NOEVENTS)
			continue;	// Ignore

		// Generate child bounds
		GUI_GenerateBounds_r (child);

		// Add to parent window bounds
		AddBoundsTo2DBounds (child->mins, child->maxs, window->mins, window->maxs);
	}
}
static void GUI_GenerateBounds (gui_t *gui)
{
	guiWindow_t	*window;
	uint32		i;

	// Clear GUI bounds
	gui->mins[0] = gui->xPos;
	gui->mins[1] = gui->yPos;
	gui->maxs[0] = gui->mins[0] + gui->width;
	gui->maxs[1] = gui->mins[1] + gui->height;

	// Generate window bounds
	for (i=0, window=gui->windowList ; i<gui->numWindows ; window++, i++) {
		if (!window->visible || window->flags & WNFL_NOEVENTS)
			continue;	// Ignore

		// Calculate window bounds
		GUI_GenerateBounds_r (window);

		// Add to GUI bounds
		AddBoundsTo2DBounds (window->mins, window->maxs, gui->mins, gui->maxs);
	}
}

/*
=============================================================================

	CURSOR COLLISION

=============================================================================
*/

/*
==================
GUI_CursorUpdate
==================
*/
static guiWindow_t *GUI_r_CursorUpdate (gui_t *gui, guiWindow_t *window)
{
	guiWindow_t	*child, *tmp, *bestChild;
	uint32		i;

	// Check for window collision
	if (!window->visible
	|| window->flags & WNFL_NOEVENTS
	|| gui->cursorX > window->maxs[0]
	|| gui->cursorY > window->maxs[1]
	|| gui->cursorX < window->mins[0]
	|| gui->cursorY < window->mins[1]) {
		GUI_QueueEvent (window->queList, &window->numQueueEvents, WNEV_MOUSE_EXIT);
		return NULL;
	}

	bestChild = window;
	GUI_QueueEvent (window->queList, &window->numQueueEvents, WNEV_MOUSE_ENTER);

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++) {
		tmp = GUI_r_CursorUpdate (gui, child);
		if (tmp)
			bestChild = tmp;
	}

	return bestChild;
}
void GUI_CursorUpdate (gui_t *gui)
{
	guiWindow_t	*window, *bestWindow;
	uint32		i;

	// Generate bounds for collision
	GUI_GenerateBounds (gui);

	if (!gui)
		return;
	gui->cursorWindow = NULL;

	// Check for GUI collision
	if (gui->cursorLocked
	|| gui->cursorX > gui->maxs[0]
	|| gui->cursorY > gui->maxs[1]
	|| gui->cursorX < gui->mins[0]
	|| gui->cursorY < gui->mins[1]) {
		GUI_QueueEvent (gui->queList, &gui->numQueueEvents, WNEV_MOUSE_EXIT);
		return;	// Ignore
	}
	GUI_QueueEvent (gui->queList, &gui->numQueueEvents, WNEV_MOUSE_ENTER);

	// Check for collision with the window children
	bestWindow = NULL;
	for (i=0, window=gui->windowList ; i<gui->numWindows ; window++, i++) {
		bestWindow = GUI_r_CursorUpdate (gui, window);
		if (bestWindow)
			gui->cursorWindow = bestWindow;
	}
}


/*
==================
GUI_MoveMouse
==================
*/
void GUI_MoveMouse (int xMove, int yMove)
{
	gui_t	*gui;

	gui = cg_guiState.inputGUI;
	if (!gui || gui->cursorLocked)
		return;

	// Move
	if (gui_mouseFilter->integer) {
		gui->cursorX = (gui->cursorX * 2 + (xMove * gui_mouseSensitivity->value)) * 0.5f;
		gui->cursorY = (gui->cursorY * 2 + (yMove * gui_mouseSensitivity->value)) * 0.5f;
	}
	else {
		gui->cursorX += xMove * gui_mouseSensitivity->value;
		gui->cursorY += yMove * gui_mouseSensitivity->value;
	}

	// Clamp
	gui->cursorX = clamp (gui->cursorX, 2, cg.glConfig.vidWidth - 2);
	gui->cursorY = clamp (gui->cursorY, 2, cg.glConfig.vidHeight - 2);

	// Update collision
	if (xMove || yMove)
		GUI_CursorUpdate (gui);
}
