/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// gui_events.c
//

#include "gui_local.h"

/*
=============================================================================

	EVENT TRIGGERING

=============================================================================
*/

/*
==================
GUI_FireActions
==================
*/
static void GUI_FireActions (gui_t *gui, guiWindow_t *window, evAction_t *actionList, uint32 numActions, wnEvent_t *event)
{
	evAction_t	*action;
	uint32		i;

	for (i=0, action=actionList ; i<numActions ; action++, i++) {
		switch (action->type) {
		case WNEA_COMMAND:
			cgi.Cbuf_AddText (action->charArgs[0]);
			break;

		case WNEA_IF:
			break;

		case WNEA_LOCAL_SOUND:
			break;

		case WNEA_LOCK_CURSOR:
			gui->cursorLocked = action->intArgs[0];
			break;

		case WNEA_RESET_TIME:
			if (gui)
				gui->time = action->intArgs[0];
			else
				window->time = action->intArgs[0];
			break;

		case WNEA_SET:
			// When visibility is modified, trigger WNEV_INIT
			break;
		case WNAE_SET_FOCUS:
			break;
		case WNAE_STOP_TRANSITIONS:
			break;
		case WNEA_TRANSITION:
			break;
		}
	}
}


/*
==================
GUI_RunEvent
==================
*/
static void GUI_RunEvent (gui_t *gui, guiWindow_t *window, wnEvent_t *eventList, uint32 numEvents, wnEventType_t type)
{
	wnEvent_t	*event;
	uint32		i;

	for (i=0, event=eventList ; i<numEvents ; event++, i++) {
		if (event->type != type)
			continue;

			switch (event->type) {
			case WNEV_NONE:
				break;

			case WNEV_TIME:
				if (event->intArgs[0] == cgi.Sys_Milliseconds ())
					GUI_FireActions (gui, window, event->actionList, event->numActions, event);
				break;

			default:
				GUI_FireActions (gui, window, event->actionList, event->numActions, event);
				break;
			}
	}
}


/*
==================
GUI_RunAllEvents
==================
*/
void GUI_r_RunFrameTimeEvents (guiWindow_t *window)
{
	guiWindow_t	*child;
	uint32		i;

	// Process all frame and time events
	GUI_RunEvent (NULL, window, window->eventList, window->numEvents, WNEV_FRAME);
	GUI_RunEvent (NULL, window, window->eventList, window->numEvents, WNEV_TIME);

	// Processed any queued event triggers
	for (i=0 ; i<window->numQueueEvents ; i++)
		GUI_RunEvent (NULL, window, window->eventList, window->numEvents, window->queList[i]);
	window->numQueueEvents = 0;

	// Process any transitions
	// FIXME...

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++)
		GUI_r_RunFrameTimeEvents (child);
}
void GUI_RunAllEvents (gui_t *gui)
{
	guiWindow_t	*window;
	uint32		i;

	if (!gui)
		return;

	// Process all frame and time events
	GUI_RunEvent (gui, NULL, gui->eventList, gui->numEvents, WNEV_FRAME);
	GUI_RunEvent (gui, NULL, gui->eventList, gui->numEvents, WNEV_TIME);

	// Processed any queued event triggers
	for (i=0 ; i<gui->numQueueEvents ; i++)
		GUI_RunEvent (gui, NULL, gui->eventList, gui->numEvents, gui->queList[i]);
	gui->numQueueEvents = 0;

	// Process any transitions
	// FIXME...

	// Process all window events
	for (i=0, window=gui->windowList ; i<gui->numWindows ; window++, i++)
		GUI_r_RunFrameTimeEvents (window);
}

/*
=============================================================================

	QUEUED EVENTS

=============================================================================
*/

/*
==================
GUI_QueueEvent
==================
*/
void GUI_QueueEvent (wnEventType_t *queList, uint32 *numQueueEvents, wnEventType_t type)
{
	if (type <= WNEV_NONE || type >= WNEV_MAX)
		return;

	// Make sure there's room
	if (*numQueueEvents+1 >= MAX_WINDOW_EVENTS)
		return;

	queList[*numQueueEvents++] = type;
}
