/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// gui_init.c
//

#include "gui_local.h"

#define BASE_MODDIRNAME		"baseq2"	// FIXME
#define EGL_GUI_VER			"EGLGUIv0.0.1"

static gui_t		cg_guiList[MAX_GUIS];
static gui_t		*cg_guiHashTree[MAX_GUI_HASH];
static uint32		cg_numGUI;

static uint32		cg_numGUIErrors;
static uint32		cg_numGUIWarnings;

cVar_t	*gui_debugBounds;
cVar_t	*gui_mouseFilter;
cVar_t	*gui_mouseSensitivity;

/*
================
GUI_HashValue

String passed to this should be lowercase.
================
*/
static uint32 GUI_HashValue (const char *name)
{
	uint32	hashValue;
	int		ch, i;

	for (i=0, hashValue=0 ; *name ; i++) {
		ch = *(name++);
		hashValue = hashValue * 33 + ch;
	}

	return (hashValue + (hashValue >> 5)) & (MAX_GUI_HASH-1);
}


/*
==================
GUI_FindGUI
==================
*/
static gui_t *GUI_FindGUI (char *guiName)
{
	gui_t	*gui, *bestMatch;
	char	tempName[MAX_QPATH];
	uint32	hashValue;

	if (!guiName)
		return NULL;

	// Make sure it's lowercase
	Q_strncpyz (tempName, guiName, sizeof (tempName));
	Q_strlwr (tempName);

	hashValue = GUI_HashValue (tempName);
	bestMatch = NULL;

	// Look for it
	for (gui=cg_guiHashTree[hashValue] ; gui ; gui=gui->hashNext) {
		if (strcmp (gui->name, tempName))
			continue;

		if (!bestMatch || gui->pathType >= bestMatch->pathType)
			bestMatch = gui;
	}

	return bestMatch;
}


/*
==================
GUI_FindWindow
==================
*/
static guiWindow_t *GUI_FindWindow (char *guiName, char *windowName)
{
	gui_t		*gui;
	guiWindow_t	*window, *bestMatch;
	char		tempName[MAX_QPATH];
	uint32		hashValue;

	if (!guiName || !windowName)
		return NULL;

	// Find the GUI
	gui = GUI_FindGUI (guiName);
	if (!gui)
		return NULL;

	// Make sure it's lowercase
	Q_strncpyz (tempName, windowName, sizeof (tempName));
	Q_strlwr (tempName);

	hashValue = GUI_HashValue (tempName);
	bestMatch = NULL;

	// Look for it
	for (window=gui->windowHashList[hashValue] ; window ; window=window->hashNext) {
		if (strcmp (window->name, tempName))
			continue;

		if (!bestMatch)
			bestMatch = window;
	}

	return bestMatch;
}

/*
=============================================================================

	GUI PARSING

=============================================================================
*/

typedef struct guiParseKey_s {
	const char	*keyWord;
	qBool		(*func) (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName);
} guiParseKey_t;

/*
================
GUI_ParseBool
================
*/
static qBool GUI_ParseBool (parse_t *ps, qBool *target)
{
	char	*str;
	qBool	tmp;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	tmp = atoi (str);
	if (tmp < 0 || tmp > 1)
		return qFalse;

	*target = tmp;
	return qTrue;
}


/*
================
GUI_ParseFloat
================
*/
static qBool GUI_ParseFloat (parse_t *ps, float *target)
{
	char	*str;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	*target = (float)atof (str);
	return qTrue;
}


/*
================
GUI_ParseFloatVec

FIXME: allow for "(#, #, #, #)" etc
================
*/
qBool GUI_ParseFloatVec (parse_t *ps, float *vec, uint32 size)
{
	char	*str;
	uint32	i;

	for (i=0 ; i<size ; i++) {
		str = Com_ParseExt (ps, qFalse);
		if (!str || !str[0])
			return qFalse;
		vec[i] = (float)atof (str);
	}

	return qTrue;
}


/*
================
GUI_ParseInt
================
*/
static qBool GUI_ParseInt (parse_t *ps, int *target)
{
	char	*str;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	*target = atoi (str);
	return qTrue;
}


/*
================
GUI_ParseString
================
*/
static char *GUI_ParseString (parse_t *ps, qBool allowNewLines)
{
	char	*str;

	str = Com_ParseExt (ps, allowNewLines);
	if (!str || !str[0])
		return "";

	return Q_strlwr (str);
}


/*
================
GUI_SkipLine
================
*/
static void GUI_SkipLine (parse_t *ps)
{
	char	*str;

	for ( ; ; ) {
		str = Com_ParseExt (ps, qFalse);
		if (!str || !str[0])
			break;
	}
}


/*
==================
GUI_SkipBlock

numOpen is the number of braces opened before calling this.
==================
*/
static void GUI_SkipBlock (parse_t *ps, int numOpen)
{
	char	*str;
	int		braceCount;

	str = Com_ParseExt (ps, qTrue);
	for (braceCount=numOpen ; braceCount>0 ; ) {
		str = Com_ParseExt (ps, qTrue);
		if (!str[0])
			return;

		else if (str[0] == '{')
			braceCount++;
		else if (str[0] == '}')
			braceCount--;
	}
}


/*
================
GUI_ParseLocation
================
*/
static char *GUI_ParseLocation (gui_t *gui, guiWindow_t *window, parse_t *ps)
{
	if (gui && gui->name[0]) {
		if (window && window->name[0])
			return Q_VarArgs ("GUI '%s' Window '%s' line #%i", gui->name, window->name, ps->currentLine);
		return Q_VarArgs ("GUI '%s' line #%i", gui->name, ps->currentLine);
	}

	return Q_VarArgs ("line #%i", ps->currentLine);
}


/*
==================
GUI_Printf
==================
*/
static void GUI_Printf (comPrint_t flags, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAX_COMPRINT];

	if (flags & PRNT_ERROR)
		cg_numGUIErrors++;
	else if (flags & PRNT_WARNING)
		cg_numGUIWarnings++;

	// Evaluate args
	va_start (argptr, fmt);
	vsnprintf (msg, sizeof (msg), fmt, argptr);
	va_end (argptr);

	// Print
	cgi.Com_Printf (flags, msg);
}


/*
==================
GUI_DevPrintf
==================
*/
static void GUI_DevPrintf (comPrint_t flags, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAX_COMPRINT];

	if (flags & PRNT_ERROR)
		cg_numGUIErrors++;
	else if (flags & PRNT_WARNING)
		cg_numGUIWarnings++;

	// Evaluate args
	va_start (argptr, fmt);
	vsnprintf (msg, sizeof (msg), fmt, argptr);
	va_end (argptr);

	// Print
	cgi.Com_DevPrintf (flags, msg);
}


/*
================
GUI_CallKeyFunc
================
*/
static qBool GUI_CallKeyFunc (gui_t *gui, guiWindow_t *window, parse_t *ps, guiParseKey_t *keys, char *token, qBool last)
{
	guiParseKey_t	*key;
	char			keyName[MAX_TOKEN_CHARS];
	char			*str;

	// Copy off a lower-case copy for faster comparisons
	Q_strncpyz (keyName, token, sizeof (keyName));
	Q_strlwr (keyName);
	str = "";

	// Cycle through the key list looking for a match
	for (key=keys ; key->keyWord ; key++) {
		// See if it matches the keyWord
		if (strcmp (key->keyWord, keyName))
			continue;

		// This is just to ignore any warnings
		if (!key->func) {
			GUI_SkipLine (ps);
			return qTrue;
		}

		// Failed to parse line
		if (!key->func (gui, window, ps, keyName)) {
			GUI_SkipLine (ps);
			return qFalse;
		}

		// Report any extra parameters
		str = GUI_ParseString (ps, qFalse);
		if (str[0]) {
			GUI_Printf (PRNT_WARNING, "ERROR: %s: unused trailing parameters after key: '%s'\n", GUI_ParseLocation (gui, window, ps), keyName);
			GUI_SkipLine (ps);
			return last;
		}

		// Parsed fine
		return qTrue;
	}

	// Not found
	if (last) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: unrecognized key: '%s' (after '%s')\n", GUI_ParseLocation (gui, window, ps), str, keyName);
		GUI_SkipLine (ps);
	}
	return last;
}

/*
=============================================================================

	CREATION

=============================================================================
*/

/*
================
GUI_NewGUIDef
================
*/
static gui_t *GUI_NewGUIDef (parse_t *ps, char *keyName, char *name)
{
	gui_t	*oldGUI;
	gui_t	*newGUI;

	// Check for space
	if (cg_numGUI+1 >= MAX_GUIS) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: Too many GUIs!\n", GUI_ParseLocation (NULL, NULL, ps));
		return NULL;
	}
	// Check name length
	if (strlen (name) >= MAX_GUI_NAMELEN) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name length is too long!\n", GUI_ParseLocation (NULL, NULL, ps), keyName);
		return NULL;
	}

	// See if another GUI exists under the same name
	oldGUI = GUI_FindGUI (name);
	if (oldGUI) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name '%s' already in use!\n", GUI_ParseLocation (NULL, NULL, ps), keyName, name);
		return NULL;
	}

	// Allocate a spot
	newGUI = &cg_guiList[cg_numGUI];
	memset (&cg_guiList[cg_numGUI], 0, sizeof (gui_t));

	// Fill values
	Q_strncpyz (newGUI->name, name, sizeof (newGUI->name));
	Q_strlwr (newGUI->name);

	// Set defaults
	Vector4Set (newGUI->generic.matColor, 1, 1, 1, 1);
	newGUI->generic.matScaleX = 1;
	newGUI->generic.matScaleY = 1;

	Vector4Set (newGUI->cursorColor, 1, 1, 1, 1);

	// Link into hash list
	newGUI->hashValue = GUI_HashValue (newGUI->name);
	newGUI->hashNext = cg_guiHashTree[newGUI->hashValue];
	cg_guiHashTree[newGUI->hashValue] = newGUI;

	cg_numGUI++;
	return newGUI;
}


/*
================
GUI_NewWindowDef
================
*/
static guiWindow_t *GUI_NewWindowDef (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName, char *name)
{
	guiWindow_t	*windowList;
	guiWindow_t	*newWindow;
	uint32		numWindows;

	// Find the target
	if (window) {
		windowList = &window->childList[window->numChildren];
		numWindows = window->numChildren;
	}
	else {
		windowList = &gui->windowList[gui->numWindows];
		numWindows = gui->numWindows;
	}

	// Check for space
	if (numWindows+1 >= MAX_GUI_WINDOWS) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: Too many windows!\n", GUI_ParseLocation (gui, window, ps));
		return NULL;
	}
	// Check name length
	if (strlen (name) >= MAX_GUI_NAMELEN) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name length is too long!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return NULL;
	}

	// Allocate a spot
	newWindow = &windowList[numWindows];

	// Fill values
	Q_strncpyz (newWindow->name, name, sizeof (newWindow->name));
	Q_strlwr (newWindow->name);

	// Set defaults
	newWindow->parent = window;
	newWindow->visible = newWindow->wasVisible = qTrue;
	newWindow->rect[0] = gui->rect[0];
	newWindow->rect[1] = gui->rect[1];
	if (newWindow->parent) {
		newWindow->rect[0] += newWindow->parent->rect[0];
		newWindow->rect[1] += newWindow->parent->rect[1];
	}

	Vector4Set (newWindow->generic.matColor, 1, 1, 1, 1);
	newWindow->generic.matScaleX = 1;
	newWindow->generic.matScaleY = 1;

	// Link it into the hash list
	newWindow->hashValue = GUI_HashValue (newWindow->name);
	newWindow->hashNext = gui->windowHashList[newWindow->hashValue];
	gui->windowHashList[newWindow->hashValue] = newWindow;

	if (window)
		window->numChildren++;
	else
		gui->numWindows++;
	return newWindow;
}


/*
==================
GUI_ParseEventAction
==================
*/
static const char	*gui_eventActionKeys[] = {
	{ "consolecmd",			},
	{ "if",					},
	{ "localsound",			},
	{ "lockcursor",			},
	{ "resettime",			},
	{ "set",				},
	{ "setfocus",			},
	{ "stoptransitions",	},
	{ "transition",			},
	{ NULL,					}
};
static uint32			gui_minActionArgs[WNEA_MAX] = {
	0,		// WNEA_NONE

	1,		// WNEA_COMMAND
	0,		// WNEA_IF
	1,		// WNEA_LOCAL_SOUND
	1,		// WNEA_LOCK_CURSOR
	1,		// WNEA_RESET_TIME
	2,		// WNEA_SET
	1,		// WNAE_SET_FOCUS
	0,		// WNAE_STOP_TRANSITIONS
	4		// WNEA_TRANSITION
};
static uint32			gui_maxActionArgs[WNEA_MAX] = {
	0,		// WNEA_NONE

	1,		// WNEA_COMMAND
	0,		// WNEA_IF
	1,		// WNEA_LOCAL_SOUND
	1,		// WNEA_LOCK_CURSOR
	1,		// WNEA_RESET_TIME
	2,		// WNEA_SET
	1,		// WNAE_SET_FOCUS
	0,		// WNAE_STOP_TRANSITIONS
	4		// WNEA_TRANSITION
};
static qBool GUI_ParseEventAction (evAction_t *actionList, uint32 *numActions, gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char		actionName[MAX_TOKEN_CHARS];
	evAction_t	*newAction;
	qBool		gotSemiColon;
	char		*str, *p;
	uint32		i;

	// Sanity check
	if (*numActions+1 >= MAX_EVENT_ACTIONS) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: too many %s event actions!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// Setup the action
	newAction = &actionList[*numActions++];
	str = GUI_ParseString (ps, qTrue);
	if (!str || !str[0]) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: unexpected EOF after %s!\n", GUI_ParseLocation (gui, window, ps), str);
		return qFalse;
	}
	if (str[0] == '}')
		return qTrue;

	Q_strncpyz (actionName, str, sizeof (actionName));
	p = strchr (actionName, ';');
	if (p)
		*p = '\0';
	for (i=WNEA_NONE+1 ; i<WNEA_MAX ; i++) {
		if (!strcmp (gui_eventActionKeys[i-1], actionName))
			break;
	}
	if (i == WNEA_MAX) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: unknown script action '%s'!\n", GUI_ParseLocation (gui, window, ps), actionName);
		return qFalse;
	}
	newAction->type = WNEA_NONE + i;

	// Parse the arguments
	gotSemiColon = qFalse;
	str = Com_ParseExt (ps, qTrue);
	for (newAction->numArgs=0 ; ; ) {
		if (!str || !str[0]) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: unexpected EOF after %s!\n", GUI_ParseLocation (gui, window, ps), str);
			return qFalse;
		}
		if (str[0] == ';') {
			gotSemiColon = qTrue;
			break;
		}
		if (strlen (str) >= MAX_ACTION_ARGLEN) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: action argument #%i length too long!\n", GUI_ParseLocation (gui, window, ps), newAction->numArgs+1);
			return qFalse;
		}

		// Check for too many arguments
		if (newAction->numArgs+1 > gui_maxActionArgs[newAction->type] || newAction->numArgs+1 > MAX_ACTION_ARGS) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: too many arguments for '%s::%s'!\n", GUI_ParseLocation (gui, window, ps), keyName, gui_eventActionKeys[newAction->type-1]);
			return qFalse;
		}

		// Add to argument list
		Q_strncpyz (newAction->charArgs[newAction->numArgs], str, sizeof (newAction->charArgs[newAction->numArgs]));

		// Trim off ';' if it's at the end of this argument
		p = strchr (newAction->charArgs[newAction->numArgs], ';');
		if (p) {
			gotSemiColon = qTrue;
			*p = '\0';
			// FIXME: check for chars after the ';'
		}
		if (strlen (newAction->charArgs[newAction->numArgs]) <= 0)
			break;
		newAction->intArgs[newAction->numArgs] = atoi (newAction->charArgs[newAction->numArgs]);
		newAction->numArgs++;
		if (gotSemiColon)
			break;

		// Next
		str = GUI_ParseString (ps, qTrue);
	}

	// Check for too few arguments
	if (newAction->numArgs < gui_minActionArgs[newAction->type]) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: too few arguments for '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// Final ';'
	if (!gotSemiColon) {
		str = GUI_ParseString (ps, qTrue);
		if (str[0] != ';') {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: expecting ';' got '%s'!\n", GUI_ParseLocation (gui, window, ps), str);
			return qFalse;
		}
		// FIXME: check for chars after the ';'
	}
	return qTrue;
}


/*
==================
GUI_NewEvent
==================
*/
static uint32			gui_minEventArgs[WNEV_MAX] = {
	0,		// WNEV_NONE

	0,		// WNEV_ACTION
	0,		// WNEV_ESC
	0,		// WNEV_FRAME
	0,		// WNEV_INIT
	0,		// WNEV_MOUSE_ENTER
	0,		// WNEV_MOUSE_EXIT
	1,		// WNEV_NAMED
	1		// WNEV_TIME
};
static uint32			gui_maxEventArgs[WNEV_MAX] = {
	0,		// WNEV_NONE

	0,		// WNEV_ACTION
	0,		// WNEV_ESC
	0,		// WNEV_FRAME
	0,		// WNEV_INIT
	0,		// WNEV_MOUSE_ENTER
	0,		// WNEV_MOUSE_EXIT
	1,		// WNEV_NAMED
	1		// WNEV_TIME
};
static qBool GUI_NewEvent (wnEventType_t type, wnEvent_t *eventList, uint32 *numEvents, gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	wnEvent_t	*newEvent;
	char		*str;

	// Sanity check
	if (*numEvents+1 >= MAX_WINDOW_EVENTS) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: too many events!\n", GUI_ParseLocation (gui, window, ps));
		return qFalse;
	}

	// Setup the event
	newEvent = &eventList[*numEvents++];
	newEvent->type = type;

	// Parse the arguments
	str = GUI_ParseString (ps, qTrue);
	for (newEvent->numArgs=0 ; ; ) {
		if (!str || !str[0]) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: unexpected EOF after %s!\n", GUI_ParseLocation (gui, window, ps), str);
			return qFalse;
		}
		if (str[0] == '{')
			break;
		if (strlen (str) >= MAX_EVENT_ARGLEN) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: event argument #%i length too long!\n", GUI_ParseLocation (gui, window, ps), newEvent->numArgs+1);
			return qFalse;
		}

		// Check for too many arguments
		if (newEvent->numArgs+1 > gui_maxEventArgs[type] || newEvent->numArgs+1 > MAX_EVENT_ARGS) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: too many arguments for '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName);
			return qFalse;
		}

		// Add to argument list
		Q_strncpyz (newEvent->charArgs[newEvent->numArgs], str, sizeof (newEvent->charArgs[newEvent->numArgs]));
		newEvent->intArgs[newEvent->numArgs] = atoi (newEvent->charArgs[newEvent->numArgs]);
		newEvent->numArgs++;

		// Next
		str = GUI_ParseString (ps, qTrue);
	}

	// Check for too few arguments
	if (newEvent->numArgs < gui_minEventArgs[type]) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: too few arguments for '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// '{' should be next
	if (str[0] != '{')
		str = GUI_ParseString (ps, qTrue);
	if (str[0] != '{') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expecting '{' after %s <args> got '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName, str);
		return qFalse;
	}

	// Parse the event actions
	for ( ; ; ) {
		if (str[0] == '}')
			break;

		if (!GUI_ParseEventAction (newEvent->actionList, &newEvent->numActions, gui, window, ps, keyName))
			return qFalse;

		if (str[0] == '}')
			break;

		str = GUI_ParseString (ps, qTrue);
	}

	// Final '}'
	if (str[0] != '}') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expecting '}' got '%s'!\n", GUI_ParseLocation (gui, window, ps), str);
		return qFalse;
	}
	return qTrue;
}

/*
=============================================================================

	BINDDEF KEYS

=============================================================================
*/

static guiParseKey_t	gui_bindDefKeys[] = {
	{ "bind",					NULL						},
	{ NULL,						NULL						}
};

/*
=============================================================================

	CHOICEDEF KEYS

=============================================================================
*/

/*
==================
choiceDef_liveUpdate
==================
*/
static qBool choiceDef_liveUpdate (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	qBool	liveUpdate;

	// Just a boolean
	if (!GUI_ParseBool (ps, &liveUpdate)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defChoice.liveUpdate = liveUpdate;
	return qTrue;
}

static guiParseKey_t	gui_choiceDefKeys[] = {
	{ "liveupdate",				&choiceDef_liveUpdate		},
	{ "choicetype",				NULL						},
	{ "currentchoice",			NULL						},
	{ "choices",				NULL						},
	{ "values",					NULL						},
	{ "cvar",					NULL						},
	{ NULL,						NULL						}
};

/*
=============================================================================

	EDITDEF KEYS

=============================================================================
*/

/*
==================
editDef_liveUpdate
==================
*/
static qBool editDef_liveUpdate (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	qBool	liveUpdate;

	// Just a boolean
	if (!GUI_ParseBool (ps, &liveUpdate)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defEdit.liveUpdate = liveUpdate;
	return qTrue;
}

static guiParseKey_t	gui_editDefKeys[] = {
	{ "liveupdate",				&editDef_liveUpdate			},
	{ "maxchars",				NULL						},
	{ "numeric",				NULL						},
	{ "wrap",					NULL						},
	{ "readonly",				NULL						},
	{ "source",					NULL						},
	{ "password",				NULL						},
	{ "cvar",					NULL						},
	{ NULL,						NULL						}
};

/*
=============================================================================

	LISTDEF KEYS

=============================================================================
*/

static guiParseKey_t	gui_listDefKeys[] = {
	{ "listname",				NULL						},
	{ "scrollbar",				NULL						},
	{ NULL,						NULL						}
};

/*
=============================================================================

	RENDERDEF KEYS

=============================================================================
*/

static guiParseKey_t	gui_renderDefKeys[] = {
	{ "lightcolor",				NULL						},
	{ "lightorigin",			NULL						},
	{ "model",					NULL						},
	{ "modelrotate",			NULL						},
	{ "vieworigin",				NULL						},
	{ "viewangles",				NULL						},
	{ NULL,						NULL						}
};

/*
=============================================================================

	SLIDERDEF KEYS

=============================================================================
*/

/*
==================
sliderDef_liveUpdate
==================
*/
static qBool sliderDef_liveUpdate (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	qBool	liveUpdate;

	if (!GUI_ParseBool (ps, &liveUpdate)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defSlider.liveUpdate = liveUpdate;
	return qTrue;
}


/*
==================
sliderDef_low
==================
*/
static qBool sliderDef_low (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	low;

	if (!GUI_ParseFloat (ps, &low)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defSlider.low = low;
	return qTrue;
}


/*
==================
sliderDef_high
==================
*/
static qBool sliderDef_high (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	high;

	if (!GUI_ParseFloat (ps, &high)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defSlider.high = high;
	return qTrue;
}


/*
==================
sliderDef_step
==================
*/
static qBool sliderDef_step (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	step;

	if (!GUI_ParseFloat (ps, &step)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defSlider.step = step;
	return qTrue;
}


/*
==================
sliderDef_vertical
==================
*/
static qBool sliderDef_vertical (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	qBool	vertical;

	if (!GUI_ParseBool (ps, &vertical)) {
		GUI_DevPrintf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defSlider.vertical = vertical;
	return qTrue;
}


/*
==================
sliderDef_thumbShader
==================
*/
static qBool sliderDef_thumbShader (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (window->defSlider.thumbName, str, sizeof (window->defSlider.thumbName));
	return qTrue;
}


/*
==================
sliderDef_cVar
==================
*/
static qBool sliderDef_cVar (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (window->defSlider.cVarName, str, sizeof (window->defSlider.cVarName));
	return qTrue;
}

static guiParseKey_t	gui_sliderDefKeys[] = {
	{ "liveupdate",				&sliderDef_liveUpdate		},
	{ "low",					&sliderDef_low				},
	{ "high",					&sliderDef_high				},
	{ "step",					&sliderDef_step				},
	{ "vertical",				&sliderDef_vertical			},
	{ "thumbshader",			&sliderDef_thumbShader		},
	{ "cvar",					&sliderDef_cVar				},
	{ NULL,						NULL						}
};

/*
=============================================================================

	TEXTDEF KEYS

=============================================================================
*/

/*
==================
textDef_font
==================
*/
static qBool textDef_font (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = GUI_ParseString (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (window->defText.font, str, sizeof (window->defText.font));
	return qTrue;
}


/*
==================
textDef_text
==================
*/
static qBool textDef_text (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = Com_ParseExt (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (window->defText.text, str, sizeof (window->defText.text));
	return qTrue;
}


/*
==================
textDef_textAlign
==================
*/
static qBool textDef_textAlign (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	int		align;

	if (!GUI_ParseInt (ps, &align)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}
	if (align < WNTA_LEFT || align > WNTA_RIGHT) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defText.align = align;
	return qTrue;
}


/*
==================
textDef_textColor
==================
*/
static qBool textDef_textColor (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, window->defText.color);
	return qTrue;
}


/*
==================
textDef_textScale
==================
*/
static qBool textDef_textScale (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	scale;

	if (!GUI_ParseFloat (ps, &scale)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	window->defText.scale = scale;
	return qTrue;
}

static guiParseKey_t	gui_textDefKeys[] = {
	{ "font",					&textDef_font				},
	{ "text",					&textDef_text				},
	{ "textalign",				&textDef_textAlign			},
	{ "textcolor",				&textDef_textColor			},
	{ "textscale",				&textDef_textScale			},
	{ NULL,						NULL						}
};

/*
=============================================================================

	WINDOWDEF KEYS

=============================================================================
*/

/*
==================
windowDef_fill
==================
*/
static qBool windowDef_fill (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, window->generic.fillColor);
	window->flags |= WNFL_FILLCOLOR;
	return qTrue;
}


/*
==================
windowDef_mat
==================
*/
static qBool windowDef_mat (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = GUI_ParseString (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (window->generic.matName, str, sizeof (window->generic.matName));
	window->flags |= WNFL_MATERIAL;
	return qTrue;
}


/*
==================
windowDef_matColor
==================
*/
static qBool windowDef_matColor (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, window->generic.matColor);
	return qTrue;
}


/*
==================
windowDef_matScaleX
==================
*/
static qBool windowDef_matScaleX (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	if (!GUI_ParseFloat (ps, &window->generic.matScaleX)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	return qTrue;
}


/*
==================
windowDef_matScaleY
==================
*/
static qBool windowDef_matScaleY (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	if (!GUI_ParseFloat (ps, &window->generic.matScaleY)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	return qTrue;
}


/*
==================
windowDef_noEvents
==================
*/
static qBool windowDef_noEvents (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	window->flags |= WNFL_NOEVENTS;
	return qTrue;
}


/*
==================
windowDef_noTime
==================
*/
static qBool windowDef_noTime (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	window->flags |= WNFL_NOTIME;
	return qTrue;
}


/*
==================
windowDef_rect
==================
*/
static qBool windowDef_rect (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}
	if (vec[2] <= 0 || vec[3] <= 0) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid %s width/height (%ix%i)!\n", GUI_ParseLocation (gui, window, ps), keyName, vec[2], vec[3]);
		return qFalse;
	}

	window->rect[0] = gui->rect[0];
	window->rect[1] = gui->rect[1];
	if (window->parent) {
		window->rect[0] += window->parent->rect[0] + vec[0];
		window->rect[1] += window->parent->rect[1] + vec[1];
	}
	else {
		window->rect[0] += vec[0];
		window->rect[1] += vec[1];
	}

	window->rect[2] = vec[2];
	window->rect[3] = vec[3];
	return qTrue;
}


/*
==================
windowDef_visible
==================
*/
static qBool windowDef_visible (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	qBool	visible;

	// Just a boolean
	if (!GUI_ParseBool (ps, &visible)) {
		GUI_DevPrintf (PRNT_WARNING, "WARNING: %s: invalid/missing '%s' value, assuming '1'!\n", GUI_ParseLocation (gui, window, ps), keyName);
		visible = qTrue;
	}

	window->visible = window->wasVisible = visible;
	return qTrue;
}

// ==========================================================================

static qBool windowDef_onAction (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_ACTION, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onESC (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_ESC, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onFrame (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_FRAME, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onInit (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_INIT, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onMouseEnter (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_MOUSE_ENTER, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onMouseExit (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_MOUSE_EXIT, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onNamedEvent (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_NAMED, window->eventList, &window->numEvents, gui, window, ps, keyName);
}
static qBool windowDef_onTime (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_TIME, window->eventList, &window->numEvents, gui, window, ps, keyName);
}

// ==========================================================================

// These are common between windows and items
guiParseKey_t			gui_windowItemSharedDefKeys[] = {
	{ "fill",					&windowDef_fill				},

	{ "mat",					&windowDef_mat				},
	{ "matcolor",				&windowDef_matColor			},
	{ "matscalex",				&windowDef_matScaleX		},
	{ "matscaley",				&windowDef_matScaleY		},

	{ "rotate",					NULL						},
	{ "shear",					NULL						},

	{ "noevents",				&windowDef_noEvents			},
	{ "notime",					&windowDef_noTime			},
	{ "rect",					&windowDef_rect				},
	{ "visible",				&windowDef_visible			},

	// Scripted events
	{ "onaction",				&windowDef_onAction			},
	{ "onesc",					&windowDef_onESC			},
	{ "onframe",				&windowDef_onFrame			},
	{ "oninit",					&windowDef_onInit			},
	{ "onmouseenter",			&windowDef_onMouseEnter		},
	{ "onmouseexit",			&windowDef_onMouseExit		},
	{ "onnamedevent",			&windowDef_onNamedEvent		},
	{ "ontime",					&windowDef_onTime			},

	{ NULL,						NULL						}
};

// ==========================================================================

/*
==================
guiDef_windowDef

windowDef "name" {
==================
*/
guiParseKey_t	gui_windowTypeKeys[];
static qBool GUI_r_WindowNameCheck (char *name, guiWindow_t *window)
{
	guiWindow_t	*child;
	uint32		i;

	// Check the current window
	if (!strcmp (name, window->name))
		return qTrue;

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++) {
		if (GUI_r_WindowNameCheck (name, child))
			return qTrue;
	}

	return qFalse;
}
static qBool GUI_WindowNameCheck (char *name, gui_t *gui)
{
	guiWindow_t	*window;
	uint32		i;

	// Recurse down the children
	for (i=0, window=gui->windowList ; i++ ; window++, i++) {
		if (GUI_r_WindowNameCheck (name, window))
			return qTrue;
	}

	return qFalse;
}
static qBool guiDef_windowDef (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char		windowName[MAX_TOKEN_CHARS];
	guiWindow_t	*newWindow, *children;
	wnEvent_t	*events;
	char		*str;

	// First is the name
	str = GUI_ParseString (ps, qTrue);
	if (!str[0] || str[0] == '{' || str[0] == '}') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expected <name> after %s, got '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName, str);
		return qFalse;
	}
	if (strlen (str) >= MAX_GUI_NAMELEN) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name length is too long!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}
	if (GUI_WindowNameCheck (str, gui)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name already in use!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}
	Q_strncpyz (windowName, str, sizeof (windowName));

	// Then a '{'
	str = Com_ParseExt (ps, qTrue);
	if (str[0] != '{') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expected '{' after %s name, got '%s'!\n", GUI_ParseLocation (gui, window, ps), keyName, str);
		return qFalse;
	}

	// Spawn a window and assign defaults
	if (window)
		newWindow = GUI_NewWindowDef (gui, window, ps, keyName, windowName);
	else
		newWindow = GUI_NewWindowDef (gui, NULL, ps, keyName, windowName);
	if (!newWindow)
		return qFalse;

	// Find the type and assign defaults
	if (strcmp (keyName, "binddef")) {
		newWindow->type = WNTP_BIND;
	}
	else if (strcmp (keyName, "choicedef")) {
		newWindow->type = WNTP_CHOICE;
		newWindow->defChoice.liveUpdate = qTrue;
	}
	else if (strcmp (keyName, "editdef")) {
		newWindow->type = WNTP_EDIT;
		newWindow->defEdit.maxChars = 128;
		newWindow->defEdit.liveUpdate = qTrue;
	}
	else if (strcmp (keyName, "listdef")) {
		newWindow->type = WNTP_LIST;
	}
	else if (strcmp (keyName, "renderdef")) {
		newWindow->type = WNTP_RENDER;
		VectorSet (newWindow->defRender.lightOrigin, -128, 0, 0);
		VectorSet (newWindow->defRender.lightColor, 1, 1, 1);
		Vector4Set(newWindow->defRender.modelColor, 1, 1, 1, 1);
		VectorSet (newWindow->defRender.viewOrigin, -128, 0, 0);
	}
	else if (strcmp (keyName, "sliderdef")) {
		newWindow->type = WNTP_SLIDER;
		newWindow->defSlider.high = 100;
		newWindow->defSlider.step = 1;
		newWindow->defSlider.liveUpdate = qTrue;
	}
	else if (strcmp (keyName, "textdef")) {
		newWindow->type = WNTP_TEXT;
		newWindow->defText.align = WNTA_LEFT;
		newWindow->defText.font[0] = '\0';	// FIXME: assign a default font
		Vector4Set (newWindow->defText.color, 1, 1, 1, 1);
		newWindow->defText.scale = 1;
	}
	else {
		newWindow->type = WNTP_GENERIC;
	}

	// Temporary storage space for children
	newWindow->numChildren = 0;
	newWindow->childList = CG_AllocTag (sizeof (guiWindow_t) * MAX_GUI_WINDOWS, qTrue, CGTAG_GUI_AUX);

	// Temporary storage space for events
	newWindow->numEvents = 0;
	newWindow->eventList = CG_AllocTag (sizeof (wnEvent_t) * MAX_WINDOW_EVENTS, qTrue, CGTAG_GUI_AUX);

	// Evaluate the windowDef's keys
	str = Com_ParseExt (ps, qTrue);
	for ( ; ; ) {
		if (str[0] == '}')
			break;	// Done

		if (!str[0]) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: unexpected EOF after %s!\n", GUI_ParseLocation (gui, newWindow, ps), str);
			return qFalse;
		}

		// Call keyname function(s)
		if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_windowItemSharedDefKeys, str, qFalse)
		&& !GUI_CallKeyFunc (gui, newWindow, ps, gui_windowTypeKeys, str, (newWindow->type == WNTP_GENERIC))) {
			switch (newWindow->type) {
			case WNTP_BIND:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_bindDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_CHOICE:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_choiceDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_EDIT:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_editDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_LIST:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_listDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_RENDER:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_renderDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_SLIDER:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_sliderDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_TEXT:
				if (!GUI_CallKeyFunc (gui, newWindow, ps, gui_textDefKeys, str, qTrue))
					return qFalse;
				break;
			case WNTP_GENERIC:
				return qFalse;
			}
		}

		str = Com_ParseExt (ps, qTrue);
	}

	// Store the child windows
	children = &newWindow->childList[0];
	if (newWindow->numChildren) {
		newWindow->childList = CG_AllocTag (sizeof (guiWindow_t) * newWindow->numChildren, qFalse, CGTAG_GUI_AUX);
		memcpy (newWindow->childList, children, sizeof (guiWindow_t) * newWindow->numChildren);
	}
	else
		newWindow->childList = NULL;
	CG_MemFree (children);

	// Store the events
	events = &newWindow->eventList[0];
	if (newWindow->numEvents) {
		newWindow->eventList = CG_AllocTag (sizeof (wnEvent_t) * newWindow->numEvents, qFalse, CGTAG_GUI_AUX);
		memcpy (newWindow->eventList, events, sizeof (wnEvent_t) * newWindow->numEvents);
	}
	else
		newWindow->eventList = NULL;
	CG_MemFree (events);
	return qTrue;
}

// Only windows have access to these (not items, not the root GUI)
guiParseKey_t			gui_windowTypeKeys[] = {
	{ "windowdef",				&guiDef_windowDef			},
	{ "binddef",				&guiDef_windowDef			},
	{ "choicedef",				&guiDef_windowDef			},
	{ "editdef",				&guiDef_windowDef			},
	{ "listdef",				&guiDef_windowDef			},
	{ "renderdef",				&guiDef_windowDef			},
	{ "sliderdef",				&guiDef_windowDef			},
	{ NULL,						NULL						}
};

/*
=============================================================================

	GUIDEF KEYS

=============================================================================
*/

/*
==================
guiDef_cursorMat
==================
*/
static qBool guiDef_cursorMat (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = GUI_ParseString (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (gui->cursorName, str, sizeof (gui->cursorName));
	return qTrue;
}


/*
==================
guiDef_cursorColor
==================
*/
static qBool guiDef_cursorColor (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, gui->cursorColor);
	return qTrue;
}


/*
==================
guiDef_cursorHeight
==================
*/
static qBool guiDef_cursorHeight (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	height;

	if (!GUI_ParseFloat (ps, &height) || height <= 0) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	gui->cursorHeight = height;
	return qTrue;
}


/*
==================
guiDef_cursorWidth
==================
*/
static qBool guiDef_cursorWidth (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	float	width;

	if (!GUI_ParseFloat (ps, &width) || width <= 0) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	gui->cursorWidth = width;
	return qTrue;
}


/*
==================
guiDef_noCursor
==================
*/
static qBool guiDef_noCursor (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	gui->flags |= GUIFL_NOCURSOR;
	return qTrue;
}


/*
==================
guiDef_menu
==================
*/
static qBool guiDef_menu (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	gui->flags |= GUIFL_MENU;
	return qTrue;
}


/*
==================
guiDef_fill
==================
*/
static qBool guiDef_fill (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, gui->generic.fillColor);
	gui->flags |= GUIFL_FILLCOLOR;
	return qTrue;
}


/*
==================
guiDef_mat
==================
*/
static qBool guiDef_mat (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char	*str;

	str = GUI_ParseString (ps, qFalse);
	if (!str || !str[0])
		return qFalse;

	Q_strncpyz (gui->generic.matName, str, sizeof (gui->generic.matName));
	gui->flags |= GUIFL_MATERIAL;
	return qTrue;
}


/*
==================
guiDef_matColor
==================
*/
static qBool guiDef_matColor (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	// FIXME: normalize color
	Vector4Copy (vec, gui->generic.matColor);
	return qTrue;
}


/*
==================
guiDef_matScaleX
==================
*/
static qBool guiDef_matScaleX (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	if (!GUI_ParseFloat (ps, &gui->generic.matScaleX)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	return qTrue;
}


/*
==================
windowDef_matScaleY
==================
*/
static qBool guiDef_matScaleY (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	if (!GUI_ParseFloat (ps, &gui->generic.matScaleY)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' value!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}

	return qTrue;
}


/*
==================
guiDef_noTime
==================
*/
static qBool guiDef_noTime (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	gui->flags |= GUIFL_NOTIME;
	return qTrue;
}


/*
==================
guiDef_rect
==================
*/
static qBool guiDef_rect (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	vec4_t	vec;

	if (!GUI_ParseFloatVec (ps, vec, 4)) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid/missing '%s' parameters!\n", GUI_ParseLocation (gui, window, ps), keyName);
		return qFalse;
	}
	if (vec[2] <= 0 || vec[3] <= 0) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: invalid %s width/height (%ix%i)!\n", GUI_ParseLocation (gui, window, ps), keyName, vec[2], vec[3]);
		return qFalse;
	}

	gui->rect[0] = vec[0];
	gui->rect[1] = vec[1];
	gui->rect[2] = vec[2];
	gui->rect[3] = vec[3];
	return qTrue;
}

// ==========================================================================

static qBool guiDef_onAction (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_ACTION, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onESC (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_ESC, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onFrame (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_FRAME, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onInit (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_INIT, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onMouseEnter (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_MOUSE_ENTER, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onMouseExit (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_MOUSE_EXIT, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onNamedEvent (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_NAMED, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}
static qBool guiDef_onTime (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	return GUI_NewEvent (WNEV_TIME, gui->eventList, &gui->numEvents, gui, window, ps, keyName);
}

// ==========================================================================

static guiParseKey_t	gui_guiDefKeys[] = {
	// Child windows
	{ "windowdef",				&guiDef_windowDef			},
	{ "binddef",				&guiDef_windowDef			},
	{ "choicedef",				&guiDef_windowDef			},
	{ "editdef",				&guiDef_windowDef			},
	{ "listdef",				&guiDef_windowDef			},
	{ "renderdef",				&guiDef_windowDef			},
	{ "sliderdef",				&guiDef_windowDef			},

	// GUI properties
	{ "cursormat",				&guiDef_cursorMat			},
	{ "cursorcolor",			&guiDef_cursorColor			},
	{ "cursorheight",			&guiDef_cursorHeight		},
	{ "cursorwidth",			&guiDef_cursorWidth			},
	{ "nocursor",				&guiDef_noCursor			},

	{ "menu",					&guiDef_menu				},

	{ "fill",					&guiDef_fill				},

	{ "mat",					&guiDef_mat					},
	{ "matcolor",				&guiDef_matColor			},
	{ "matscalex",				&guiDef_matScaleX			},
	{ "matscaley",				&guiDef_matScaleY			},

	{ "rotate",					NULL						},
	{ "shear",					NULL						},

	{ "notime",					&guiDef_noTime				},
	{ "rect",					&guiDef_rect				},

	// Scripted events
	{ "onaction",				&guiDef_onAction			},
	{ "onesc",					&guiDef_onESC				},
	{ "onframe",				&guiDef_onFrame				},
	{ "oninit",					&guiDef_onInit				},
	{ "onmouseenter",			&guiDef_onMouseEnter		},
	{ "onmouseexit",			&guiDef_onMouseExit			},
	{ "onnamedevent",			&guiDef_onNamedEvent		},
	{ "ontime",					&guiDef_onTime				},

	{ NULL,						NULL						}
};

// ==========================================================================

/*
==================
base_guiDef
==================
*/
static qBool base_guiDef (gui_t *gui, guiWindow_t *window, parse_t *ps, char *keyName)
{
	char		guiName[MAX_TOKEN_CHARS];
	gui_t		*newGUI;
	guiWindow_t	*children;
	wnEvent_t	*events;
	char		*str;

	// First is the name
	str = GUI_ParseString (ps, qTrue);
	if (!str[0] || str[0] == '{' || str[0] == '}') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expected <name> after %s, got '%s'!\n", GUI_ParseLocation (NULL, NULL, ps), keyName, str);
		return qFalse;
	}
	if (strlen (str) >= MAX_GUI_NAMELEN) {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: %s name length is too long!\n", GUI_ParseLocation (NULL, NULL, ps), keyName);
		return qFalse;
	}
	Q_strncpyz (guiName, str, sizeof (guiName));

	// Then a '{'
	str = Com_ParseExt (ps, qTrue);
	if (str[0] != '{') {
		GUI_Printf (PRNT_ERROR, "ERROR: %s: expected '{' after %s name, got '%s'!\n", GUI_ParseLocation (NULL, NULL, ps), keyName, str);
		return qFalse;
	}

	// Spawn a window and assign defaults
	newGUI = GUI_NewGUIDef (ps, keyName, guiName);
	if (!newGUI)
		return qFalse;

	// Temporary storage space for children
	newGUI->numWindows = 0;
	newGUI->windowList = CG_AllocTag (sizeof (guiWindow_t) * MAX_GUI_WINDOWS, qTrue, CGTAG_GUI_AUX);

	// Temporary storage space for events
	newGUI->numEvents = 0;
	newGUI->eventList = CG_AllocTag (sizeof (wnEvent_t) * MAX_WINDOW_EVENTS, qTrue, CGTAG_GUI_AUX);

	// Evaluate the windowDef's keys
	str = Com_ParseExt (ps, qTrue);
	for ( ; ; ) {
		if (str[0] == '}')
			break;	// Done

		if (!str[0]) {
			GUI_Printf (PRNT_ERROR, "ERROR: %s: unexpected EOF after %s!\n", GUI_ParseLocation (newGUI, NULL, ps), str);
			return qFalse;
		}

		// Call keyname function(s)
		if (!GUI_CallKeyFunc (newGUI, NULL, ps, gui_guiDefKeys, str, qTrue))
			return qFalse;

		str = Com_ParseExt (ps, qTrue);
	}

	// Store the child windows
	children = newGUI->windowList;
	if (newGUI->numWindows) {
		newGUI->windowList = CG_AllocTag (sizeof (guiWindow_t) * newGUI->numWindows, qFalse, CGTAG_GUI_AUX);
		memcpy (newGUI->windowList, children, sizeof (guiWindow_t) * newGUI->numWindows);
	}
	else
		newGUI->windowList = NULL;
	CG_MemFree (children);

	// Store the events
	events = newGUI->eventList;
	if (newGUI->numEvents) {
		newGUI->eventList = CG_AllocTag (sizeof (wnEvent_t) * newGUI->numEvents, qFalse, CGTAG_GUI_AUX);
		memcpy (newGUI->eventList, events, sizeof (wnEvent_t) * newGUI->numEvents);
	}
	else
		newGUI->eventList = NULL;
	CG_MemFree (events);

	// Move all items in the AUX tag to the regular tag, as this GUI is valid!
	CG_ChangeTag (CGTAG_GUI_AUX, CGTAG_GUI);
	return qTrue;
}

/*
=============================================================================

	GUI FILE LOADING

=============================================================================
*/

static guiParseKey_t	gui_baseKeys[] = {
	{ "guidef",					&base_guiDef				},
	{ NULL,						NULL						}
};

// ==========================================================================

/*
==================
GUI_ParseScript
==================
*/
static void GUI_ParseScript (char *fixedName, guiPathType_t pathType)
{
	char	*fileBuffer, *buf;
	char	*token;
	int		fileLen;
	int		numSlashes;
	int		i;
	parse_t	*ps;

	if (!fixedName)
		return;

	// Check for recursion
	for (numSlashes=0, i=0 ; fixedName[i] ; i++) {
		if (fixedName[i] == '/')
			numSlashes++;
	}
	if (numSlashes > 1)
		return;

	// Load the file
	fileLen = cgi.FS_LoadFile (fixedName, (void **)&fileBuffer); 
	if (!fileBuffer || fileLen <= 0) {
		GUI_Printf (PRNT_WARNING, "...WARNING: can't load '%s' -- %s\n", fixedName, !fileBuffer ? "unable to open" : "no data");
		return;
	}

	GUI_Printf (0, "...loading '%s'\n", fixedName);

	// Copy the buffer, and make certain it's newline and null terminated
	buf = (char *)CG_MemAlloc (fileLen+2, qFalse);
	memcpy (buf, fileBuffer, fileLen);
	buf[fileLen] = '\n';
	buf[fileLen+1] = '\0';

	// Don't need this anymore
	CG_FS_FreeFile (fileBuffer);
	fileBuffer = buf;

	// Start parsing
	ps = Com_BeginParseSession (&fileBuffer);

	// Pull out the version and check it
	token = Com_ParseExt (ps, qTrue);
	if (!token || !token[0]) {
		GUI_Printf (PRNT_ERROR, "ERROR: No version string at beginning of file, ignoring!\n");
		Com_EndParseSession (ps);
		CG_MemFree (buf);
		CG_FreeTag (CGTAG_GUI_AUX);
		return;
	}
	if (strcmp (token, EGL_GUI_VER)) {
		GUI_Printf (PRNT_ERROR, "ERROR: Version mismatch ('%s' != '%s'), ignoring!\n", token, EGL_GUI_VER);
		Com_EndParseSession (ps);
		CG_MemFree (buf);
		CG_FreeTag (CGTAG_GUI_AUX);
		return;
	}

	// Evaluate the guiDef's keys
	token = Com_ParseExt (ps, qTrue);
	for ( ; ; ) {
		if (!token || !token[0])
			break;

		// Call keyname function(s)
		if (!GUI_CallKeyFunc (NULL, NULL, ps, gui_baseKeys, token, qTrue)) {
			// Failed! Remove this and skip to the next guiDef
			CG_FreeTag (CGTAG_GUI_AUX);
			memset (&cg_guiList[--cg_numGUI], 0, sizeof (gui_t));

			GUI_SkipBlock (ps, (token[0] == '{') ? 1 : 0);
			break;
		}

		token = Com_ParseExt (ps, qTrue);
	}

	// Done
	CG_MemFree (buf);
	CG_FreeTag (CGTAG_GUI_AUX);
	Com_EndParseSession (ps);
}

/*
=============================================================================

	REGISTRATION

=============================================================================
*/

/*
==================
GUI_TouchGUI
==================
*/
static void GUI_CantFindMedia (gui_t *gui, guiWindow_t *window, char *item, char *name)
{
	if (gui && gui->name[0]) {
		if (window && window->name[0])
			GUI_Printf (PRNT_WARNING, "WARNING: GUI '%s' Window '%s': can't find %s '%s'", gui->name, window->name, item, name);
		else
			GUI_Printf (PRNT_WARNING, "WARNING: GUI '%s': can't find %s '%s'", gui->name, item, name);
		return;
	}

	GUI_Printf (PRNT_WARNING, "WARNING: can't find %s '%s'", item, name);
}
static void GUI_TouchGUI_r (gui_t *gui, guiWindow_t *window)
{
	guiWindow_t	*child;
	uint32		i;

	// Register all window media
	if (window->flags & WNFL_MATERIAL)
		window->generic.material = cgi.R_RegisterPic (window->generic.matName);

	switch (window->type) {
	case WNTP_BIND:
		break;

	case WNTP_CHOICE:
		window->defChoice.cVarPtr = cgi.Cvar_Exists (window->defChoice.cVarName);
		if (!window->defChoice.cVarPtr)
			GUI_CantFindMedia (gui, window, "cVar", window->defChoice.cVarName);
		break;

	case WNTP_EDIT:
		window->defEdit.cVarPtr = cgi.Cvar_Exists (window->defEdit.cVarName);
		if (!window->defEdit.cVarPtr)
			GUI_CantFindMedia (gui, window, "cVar", window->defEdit.cVarName);
		// FIXME: load the source file here
		break;

	case WNTP_LIST:
		break;

	case WNTP_RENDER:
		window->defRender.modelPtr = cgi.R_RegisterModel (window->defRender.modelName);
		if (!window->defRender.modelPtr)
			GUI_CantFindMedia (gui, window, "modelName", window->defRender.modelName);

		window->defRender.modelMatPtr = cgi.R_RegisterSkin (window->defRender.modelMatName);
		if (!window->defRender.modelMatPtr)
			GUI_CantFindMedia (gui, window, "modelMat", window->defRender.modelMatName);
		break;

	case WNTP_SLIDER:
		window->defSlider.cVarPtr = cgi.Cvar_Exists (window->defSlider.cVarName);
		if (!window->defSlider.cVarPtr)
			GUI_CantFindMedia (gui, window, "cVar", window->defSlider.cVarName);

		window->defSlider.thumbShader = cgi.R_RegisterPic (window->defSlider.thumbName);
		if (!window->defSlider.thumbShader)
			GUI_CantFindMedia (gui, window, "thumbShader", window->defSlider.thumbName);
		break;

	case WNTP_TEXT:
		// FIXME: load the font here
		break;
	}

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++)
		GUI_TouchGUI_r (gui, child);
}
void GUI_TouchGUI (gui_t *gui)
{
	guiWindow_t	*window;
	uint32		i;

	// Register all GUI media
	if (!(gui->flags & GUIFL_NOCURSOR) && gui->cursorName[0])
		gui->cursorShader = cgi.R_RegisterPic (gui->cursorName);
	if (gui->flags & GUIFL_MATERIAL)
		gui->generic.material = cgi.R_RegisterPic (gui->generic.matName);

	// Touch each window
	for (i=0, window=gui->windowList ; i<gui->numWindows ; window++, i++)
		GUI_TouchGUI_r (gui, window);
}


/*
==================
GUI_RegisterGUI
==================
*/
gui_t *GUI_RegisterGUI (char *name)
{
	gui_t	*gui;

	// Find it
	gui = GUI_FindGUI (name);
	if (!gui)
		return NULL;

	// Touch it
	GUI_TouchGUI (gui);
	return gui;
}


/*
==================
GUI_RegisterMedia

FIXME: Needs a registration sequence, so that GUI sounds/images
will be registered on map load for constantly-used GUIs, and only
on use for the rest.
==================
*/
void GUI_RegisterMedia (void)
{
	gui_t		*gui;
	uint32		i;

	// Register media specific to GUI windows
	for (i=0, gui=cg_guiList ; i<cg_numGUI ; gui++, i++)
		GUI_TouchGUI (gui);
}

/*
=============================================================================

	INIT / SHUTDOWN

=============================================================================
*/

static void *cmd_gui_list;
static void	*cmd_gui_restart;
static void *cmd_gui_test;

/*
==================
GUI_List_f_r1
==================
*/
static GUI_List_f_r1 (guiWindow_t *window, uint32 depth)
{
	guiWindow_t	*child;
	uint32		i, j;

	// Window title
	for (j=0 ; j<depth ; j++)
		Com_Printf (0, " ");
	Com_Printf (0, "Window \"%s\" (%i children):\n", window->name, window->numChildren);

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++)
		GUI_List_f_r1 (child, depth+1);
}
static void GUI_List_f (void)
{
	gui_t		*gui;
	guiWindow_t	*window;
	uint32		i, j;

	Com_Printf (0, "GUI List:\n");
	for (i=0, gui=cg_guiList ; i<cg_numGUI ; gui++, i++) {
		Com_Printf (0, "GUI #%i %s (%i windows)\n", i, gui->name, gui->numWindows);
		for (j=0, window=gui->windowList ; j<gui->numWindows ; window++, j++)
			GUI_List_f_r1 (window, 1);
	}
}


/*
==================
GUI_Restart_f
==================
*/
static void GUI_Restart_f (void)
{
	GUI_Shutdown ();
	GUI_Init ();
}


/*
==================
GUI_Test_f
==================
*/
static void GUI_Test_f (void)
{
	gui_t	*gui;

	if (cgi.Cmd_Argc () < 2) {
		Com_Printf (0, "Usage: gui_test <GUI name>\n");
		return;
	}

	// Find it and touch it
	gui = GUI_RegisterGUI (cgi.Cmd_Argv (1));
	if (!gui) {
		Com_Printf (0, "Not found...\n");
		return;
	}

	// Open it
	GUI_OpenGUI (gui);
}


/*
==================
GUI_Init
==================
*/
void GUI_Init (void)
{
	char			*guiList[MAX_GUIS];
	char			fixedName[MAX_QPATH];
	uint32			numGUI, i;
	guiPathType_t	pathType;
	char			*name;

	Com_Printf (0, "\n---------- GUI Initialization ----------\n");
	Com_Printf (0, "Version: %s\n", EGL_GUI_VER);

	// Clear lists
	cg_numGUI = 0;
	memset (&cg_guiList, 0, sizeof (gui_t) * MAX_GUIS);
	memset (cg_guiHashTree, 0, sizeof (gui_t *) * MAX_GUI_HASH);

	// Console commands
	cmd_gui_list	= cgi.Cmd_AddCommand ("gui_list",		GUI_List_f,		"List GUIs in memory");
	cmd_gui_restart	= cgi.Cmd_AddCommand ("gui_restart",	GUI_Restart_f,	"Restart the GUI system");
	cmd_gui_test	= cgi.Cmd_AddCommand ("gui_test",		GUI_Test_f,		"Test the specified GUI");

	// Cvars
	gui_debugBounds			= cgi.Cvar_Register ("gui_debugBounds",				"0",		0);
	gui_mouseFilter			= cgi.Cvar_Register ("gui_mouseFilter",				"1",		CVAR_ARCHIVE);
	gui_mouseSensitivity	= cgi.Cvar_Register ("gui_mouseSensitivity",		"2",		CVAR_ARCHIVE);

	// Load scripts
	cg_numGUIWarnings = 0;
	cg_numGUIErrors = 0;
	numGUI = cgi.FS_FindFiles ("guis", "*guis/*.gui", "gui", guiList, MAX_GUIS, qTrue, qFalse);
	for (i=0 ; i<numGUI ; i++) {
		// Fix the path
		Com_NormalizePath (fixedName, sizeof (fixedName), guiList[i]);

		// Skip the path
		name = strstr (fixedName, "/guis/");
		if (!name)
			continue;
		name++;	// Skip the initial '/'

		// Base dir GUI?
		if (strstr (guiList[i], BASE_MODDIRNAME "/"))
			pathType = GUIPT_BASEDIR;
		else
			pathType = GUIPT_MODDIR;

		// Parse it
		GUI_ParseScript (name, pathType);
	}
	CG_FS_FreeFileList (guiList, numGUI);

	// Free auxillery (scratch space) tags
	CG_FreeTag (CGTAG_GUI_AUX);

	Com_Printf (0, "GUI - %i error(s), %i warning(s)\n", cg_numGUIErrors, cg_numGUIWarnings);
	Com_Printf (0, "----------------------------------------\n");
}


/*
==================
GUI_Shutdown
==================
*/
void GUI_Shutdown (void)
{
	// Close all open GUIs
	GUI_CloseAllGUIs ();

	// Remove commands
	cgi.Cmd_RemoveCommand ("gui_list", cmd_gui_list);
	cgi.Cmd_RemoveCommand ("gui_restart", cmd_gui_restart);
	cgi.Cmd_RemoveCommand ("gui_test", cmd_gui_test);

	// Release all used memory
	CG_FreeTag (CGTAG_GUI);
	CG_FreeTag (CGTAG_GUI_AUX);
}
