/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// gui_local.h
//

#include "cg_local.h"

#define MAX_GUIS			512			// Maximum parsed GUIs in memory
#define MAX_GUI_DEPTH		32			// Maximum GUIs open at one time
#define MAX_GUI_HASH		64
#define MAX_GUI_NAMELEN		64			// Maximum GUI name length
#define MAX_GUI_WINDOWS		64			// Maximum GUI Windows (per GUI and per Window)

#define MAX_WINDOW_EVENTS	32			// Maximum number of events per window

#define MAX_EVENT_ARGS		2			// Maximum number of event arguments (per event)
#define MAX_EVENT_ARGLEN	64			// Maximum length of an event argument (per event)
#define MAX_EVENT_ACTIONS	16			// Maximum number of actions per event (per GUI and per Window)

#define MAX_ACTION_ARGS		4			// Maximum number of action arguments (per action)
#define MAX_ACTION_ARGLEN	64			// Maximum length of an action argument (per argument)

/*
=============================================================================

	GUI WINDOW EVENTS

=============================================================================
*/

//
// Event types
//
typedef enum wnEventType_s {
	WNEV_NONE,

	WNEV_ACTION,
	WNEV_ESC,
	WNEV_FRAME,
	WNEV_INIT,
	WNEV_MOUSE_ENTER,
	WNEV_MOUSE_EXIT,
	WNEV_NAMED,
	WNEV_TIME,

	WNEV_MAX
} wnEventType_t;

//
// Action types
//
typedef enum evActionType_s {
	WNEA_NONE,

	WNEA_COMMAND,
	WNEA_IF,
	WNEA_LOCAL_SOUND,
	WNEA_LOCK_CURSOR,
	WNEA_RESET_TIME,
	WNEA_SET,
	WNAE_SET_FOCUS,
	WNAE_STOP_TRANSITIONS,
	WNEA_TRANSITION,

	WNEA_MAX
} evActionType_t;

//
// Window/GUI event actions
//
typedef struct evAction_s {
	evActionType_t		type;

	uint32				numArgs;
	char				charArgs[MAX_ACTION_ARGS][MAX_ACTION_ARGLEN];
	int					intArgs[MAX_ACTION_ARGS];
} evAction_t;

//
// Window/GUI events
//
typedef struct wnEvent_s {
	wnEventType_t		type;

	uint32				numArgs;
	char				charArgs[MAX_EVENT_ARGS][MAX_EVENT_ARGLEN];
	int					intArgs[MAX_EVENT_ARGS];

	uint32				numActions;
	evAction_t			actionList[MAX_EVENT_ACTIONS];
} wnEvent_t;

/*
=============================================================================

	GUI WINDOWS

	Must be contained in a guiDef.
	Windows contain children Windows.
=============================================================================
*/

//
// Window types
//
typedef enum wnType_s {
	WNTP_BIND,
	WNTP_CHOICE,
	WNTP_EDIT,
	WNTP_LIST,
	WNTP_RENDER,
	WNTP_SLIDER,
	WNTP_TEXT,

	WNTP_GENERIC,

	WNTP_MAX
} wnType_t;

//
// Window text alignment
//
typedef enum wnTextAlign_s {
	WNTA_LEFT,
	WNTA_CENTER,
	WNTA_RIGHT,
} wnTextAlign_t;

//
// Window flags
//
typedef enum wnFlags_s {
	WNFL_FILLCOLOR		= 1 << 0,
	WNFL_MATERIAL		= 1 << 1,
	WNFL_NOEVENTS		= 1 << 2,
	WNFL_NOTIME			= 1 << 3
} wnFlags_t;

//
// Window types
//
typedef struct guiBindDef_s {
	char				*bind;
} guiBindDef_t;

typedef struct guiChoiceDef_s {
	int					choiceType;
	int					currentChoice;

	int					numChoices;
	char				**choices;
	char				**values;

	qBool				liveUpdate;
	char				*cVarName;
	cVar_t				*cVarPtr;
} guiChoiceDef_t;

typedef struct guiEditDef_s {
	int					maxChars;
	qBool				numeric;
	qBool				wrap;
	qBool				readOnly;
	char				*source;
	qBool				password;

	qBool				liveUpdate;
	char				*cVarName;
	cVar_t				*cVarPtr;
} guiEditDef_t;

typedef struct guiListDef_s {
	char				listName[MAX_GUI_NAMELEN];
	qBool				scrollBar;
} guiListDef_t;

typedef struct guiRenderDef_s {
	vec3_t				lightColor;
	vec3_t				lightOrigin;

	vec4_t				modelColor;
	char				modelName[MAX_QPATH];
	struct model_s		*modelPtr;
	char				modelMatName[MAX_QPATH];
	struct shader_s		*modelMatPtr;
	vec3_t				modelOrigin;
	vec3_t				modelRotate;

	vec3_t				viewOrigin;
	vec3_t				viewAngles;
} guiRenderDef_t;

typedef struct guiSliderDef_s {
	float				low;
	float				high;
	float				step;
	qBool				vertical;

	char				thumbName[MAX_QPATH];
	struct shader_s		*thumbShader;

	qBool				liveUpdate;
	char				*cVarName;
	cVar_t				*cVarPtr;
} guiSliderDef_t;

typedef struct guiTextDef_s {
	char				font[MAX_QPATH];
	char				text[MAX_TOKEN_CHARS];
	wnTextAlign_t		align;
	vec4_t				color;
	float				scale;
} guiTextDef_t;

typedef struct guiWnGeneric_s {
	vec4_t				fillColor;
	vec4_t				matColor;
	char				matName[MAX_QPATH];
	struct shader_s		*material;
	float				matScaleX;
	float				matScaleY;

	float				rotation;
} guiWnGeneric_t;

typedef struct guiWindow_s {
	// Constant information
	wnType_t			type;
	wnFlags_t			flags;
	char				name[MAX_GUI_NAMELEN];

	float				rect[4];
	qBool				wasVisible;					// Never changes, used for resetting visibility state

	guiWnGeneric_t		generic;
	guiBindDef_t		defBind;
	guiChoiceDef_t		defChoice;
	guiEditDef_t		defEdit;
	guiListDef_t		defList;
	guiRenderDef_t		defRender;
	guiSliderDef_t		defSlider;
	guiTextDef_t		defText;

	// Event trigger queuing
	uint32				numQueueEvents;
	wnEventType_t		queList[MAX_WINDOW_EVENTS];

	// Event handling
	uint32				numEvents;
	wnEvent_t			*eventList;

	// Child/parent windows
	struct guiWindow_s	*parent;
	uint32				numChildren;
	struct guiWindow_s	*childList;

	// Dynamic information
	float				xPos, yPos;					// What all of the windows will be relative to
	float				width, height;

	uint32				time;						// Time in milliseconds, used for transition effects
	qBool				visible;					// Toggle Window visibility

	vec2_t				mins, maxs;

	uint32				hashValue;
	struct guiWindow_s	*hashNext;
} guiWindow_t;

/*
=============================================================================

	GUI

	Contains purely global GUI settings and the windowDefs.
=============================================================================
*/

typedef enum guiFlags_s {
	GUIFL_FILLCOLOR		= 1 << 0,
	GUIFL_MATERIAL		= 1 << 1,
	GUIFL_MENU			= 1 << 3,
	GUIFL_NOCURSOR		= 1 << 4,
	GUIFL_NOTIME		= 1 << 5,
} guiFlags_t;

typedef enum guiPathType_s {
	GUIPT_BASEDIR,
	GUIPT_MODDIR,
} guiPathType_t;

typedef struct gui_s {
	// Constant information
	guiPathType_t		pathType;
	guiFlags_t			flags;
	char				name[MAX_GUI_NAMELEN];

	float				rect[4];					// Never changes

	guiWnGeneric_t		generic;

	// Cursor information
	char				cursorName[MAX_QPATH];
	struct shader_s		*cursorShader;
	vec4_t				cursorColor;

	qBool				cursorLocked;
	float				cursorWidth, cursorHeight;
	float				cursorX, cursorY;
	guiWindow_t			*cursorWindow;

	// Event trigger queuing
	uint32				numQueueEvents;
	wnEventType_t		queList[MAX_WINDOW_EVENTS];

	// Event handling
	uint32				numEvents;
	wnEvent_t			*eventList;

	// Child windows
	uint32				numWindows;
	guiWindow_t			*windowList;
	guiWindow_t			*windowHashList[MAX_GUI_HASH];

	// Dynamic information
	float				xPos, yPos;					// What all of the Windows will be relative to
	float				width, height;

	uint32				time;						// Time in milliseconds, used for transition effects

	vec2_t				mins, maxs;

	uint32				hashValue;
	struct gui_s		*hashNext;
} gui_t;

/*
=============================================================================

	GUI STATE

=============================================================================
*/

typedef struct guiState_s {
	// GUI information
	gui_t				*inputGUI;					// GUI that will receive input
	uint32				guiDepth;
	gui_t				*guiLayers[MAX_GUI_DEPTH];	// Currently opened GUIs
} guiState_t;

extern guiState_t	cg_guiState;

/*
=============================================================================

	CVARS

=============================================================================
*/

extern cVar_t	*gui_debugBounds;
extern cVar_t	*gui_mouseFilter;
extern cVar_t	*gui_mouseSensitivity;

/*
=============================================================================

	FUNCTION PROTOTYPES

=============================================================================
*/

//
// gui_cursor.c
//
void		GUI_CursorUpdate (gui_t *gui);

//
// gui_events.c
//
void		GUI_RunAllEvents (gui_t *gui);
void		GUI_QueueEvent (wnEventType_t *queList, uint32 *numQueueEvents, wnEventType_t type);
