/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// gui_main.c
//

#include "gui_local.h"

guiState_t	cg_guiState;

/*
=============================================================================

	GUI HANDLING

=============================================================================
*/

/*
==================
GUI_OpenGUI
==================
*/
void GUI_OpenGUI (gui_t *gui)
{
	uint32		i;

	if (!gui)
		return;

	// Make sure it's not already open once
	for (i=0 ; i<cg_guiState.guiDepth ; i++) {
		if (gui == cg_guiState.guiLayers[i])
			break;
	}
	if (i != cg_guiState.guiDepth) {
		Com_Printf (0, "GUI_OpenGUI: GUI already open!\n", gui->name);
		return;
	}

	// Sanity check
	if (cg_guiState.guiDepth+1 >= MAX_GUI_DEPTH)
		Com_Error (ERR_FATAL, "GUI_OpenGUI: MAX_GUI_DEPTH");

	// Add to the list
	cg_guiState.guiLayers[cg_guiState.guiDepth] = gui;
	cg_guiState.guiDepth++;

	// Check if this will be the key grabber
	if (!(gui->flags & GUIFL_NOCURSOR)) {
		if (cg_guiState.inputGUI) {
			if (gui->flags & GUIFL_MENU)
				cg_guiState.inputGUI = gui;
		}
		else {
			cg_guiState.inputGUI = gui;
		}
	}

	// Queue the init event
	GUI_QueueEvent (gui->queList, &gui->numQueueEvents, WNEV_INIT);
}


/*
==================
GUI_CloseGUI
==================
*/
static void GUI_CloseGUI_r (guiWindow_t *window)
{
	guiWindow_t	*child;
	uint32		i;

	// Reset window visibility
	window->visible = window->wasVisible;

	// Recurse down the children
	for (i=0, child=window->childList ; i<window->numChildren ; child++, i++)
		GUI_CloseGUI_r (child);
}
void GUI_CloseGUI (gui_t *gui)
{
	gui_t		*bestGUI;
	guiWindow_t	*window;
	qBool		found;
	uint32		i;

	if (!gui)
		return;

	// Make certain it's open first
	for (i=0 ; i<cg_guiState.guiDepth ; i++) {
		if (gui == cg_guiState.guiLayers[i])
			break;
	}
	if (i == cg_guiState.guiDepth) {
		Com_Printf (0, "GUI_CloseGUI: GUI %s was not open!\n", gui->name);
		return;
	}

	// Remove from the open GUIs list
	found = qFalse;
	for (i=0 ; i<cg_guiState.guiDepth ; i++) {
		if (found)
			cg_guiState.guiLayers[i-1] = cg_guiState.guiLayers[i];
		else if (cg_guiState.guiLayers[i] == gui)
			found = qTrue;
	}
	if (found)
		cg_guiState.guiLayers[--cg_guiState.guiDepth] = NULL;

	// Return visibility states to how they originally were
	gui->cursorWindow = NULL;
	for (i=0, window=gui->windowList ; i<gui->numWindows ; window++, i++)
		GUI_CloseGUI_r (window);

	// Find the next best input GUI
	bestGUI = cg_guiState.guiLayers[0];
	for (i=1, gui=cg_guiState.guiLayers[1] ; i<cg_guiState.guiDepth ; gui++, i++) {
		if (gui->flags & GUIFL_MENU || !(bestGUI->flags & GUIFL_MENU))
			bestGUI = gui;
	}
	cg_guiState.inputGUI = bestGUI;
}


/*
==================
GUI_CloseAllGUIs
==================
*/
void GUI_CloseAllGUIs (void)
{
	gui_t	*gui;
	uint32	i;

	// Simply close all open layers
	for (i=0, gui=cg_guiState.guiLayers[0] ; i<cg_guiState.guiDepth ; gui++, i++)
		GUI_CloseGUI (gui);
}
